import sys
import unittest

from satella.coding import SelfClosingGenerator, hint_with_length, chain
from satella.coding.sequences import smart_enumerate, ConstruableIterator, walk, \
    IteratorListAdapter, is_empty, ListWrapperIterator


class TestIterators(unittest.TestCase):

    def test_list_wrapper_iterator(self):
        a = {'count': 0}

        def iterate():
            nonlocal a

            a['count'] += 1
            yield 1
            a['count'] += 1
            yield 2
            a['count'] += 1
            yield 3

        lwe = ListWrapperIterator(iterate())
        self.assertEqual(list(iter(lwe)), [1, 2, 3])
        self.assertEqual(a['count'], 3)
        self.assertEqual(list(iter(lwe)), [1, 2, 3])
        self.assertEqual(a['count'], 3)
        lwe2 = ListWrapperIterator(iterate())
        self.assertEqual(lwe2[1:2], [2])
        self.assertEqual(a['count'], 5)
        self.assertEqual(lwe2[2], 3)
        self.assertEqual(a['count'], 6)
        self.assertEqual(lwe2[0:0], [])
        self.assertRaises(IndexError, lambda: lwe2[4])

    def test_is_empty_not_exhaust(self):
        def generator():
            yield 1
            raise ValueError()

        self.assertFalse(is_empty(generator(), exhaust=False))
        self.assertTrue(is_empty([], exhaust=False))

    def test_generator_list_adapter(self):
        gla = IteratorListAdapter(range(10))
        self.assertEqual(next(gla), 0)
        self.assertEqual(gla[0], 1)
        self.assertEqual(len(gla), 9)
        self.assertEqual(list(gla), list(range(1, 10)))

    def test_walk(self):
        a = [[1, 2, 3], 4, 5, 6, [7, 8, 9]]
        b = walk(a, lambda x: x if isinstance(x, list) else None, leaves_only=True)
        self.assertEqual(list(b), [1, 2, 3, 4, 5, 6, 7, 8, 9])

    def test_construable_iterator(self):
        a = ConstruableIterator([1, 2, 3])
        c = []
        for b in a:
            if b == 2:
                a.add(5)
            c.append(b)
        self.assertEqual(c, [1, 2, 3, 5])

    def test_chain(self):
        a = chain(1, 2, [3, 4, 5], 6, (i for i in range(2)))
        a = list(a)
        self.assertEqual(a, [1, 2, 3, 4, 5, 6, 0, 1])

    def test_smart_enumerate(self):
        a = [(1, 2), (3, 4), (5, 6)]
        b = list(smart_enumerate(a))
        self.assertEqual([(0, 1, 2), (1, 3, 4), (2, 5, 6)], b)

    def test_hint_with_length(self):
        def generator():
            yield from range(1000)

        g = hint_with_length(generator, 1000)()
        self.assertEqual(g.__length_hint__(), 1000)

    @unittest.skipUnless(sys.implementation.name == 'cpython', 'Not CPython, this needs deterministic GC')
    def test_self_closing_generator_function(self):
        a = {'done': False}

        def generator():
            for i in range(5):
                yield i
            a['done'] = True

        for b in SelfClosingGenerator(generator)():
            if b == 2:
                break

        self.assertTrue(a['done'])

    def test_self_closing_generator_function(self):
        a = {'done': False}

        def generator():
            for i in range(5):
                yield i
            a['done'] = True

        with SelfClosingGenerator(generator)() as gen:
            for b in gen:
                if b == 2:
                    break

        self.assertTrue(a['done'])



import logging
import unittest
from satella.imports import import_class
import subprocess
import warnings

logger = logging.getLogger(__name__)


class TestImports(unittest.TestCase):
    def test_imports(self):
        import tests.test_imports.importa
        with warnings.catch_warnings(record=True) as warns:
            tests.test_imports.importa.do_import()
        self.assertGreater(len(warns), 0)

        # this as well checks for the namespace's pollution
        self.assertEqual(set(tests.test_imports.importa.importb.__all__),
                         {'logger', 'sub', 'logging', 'add'})

        self.assertEqual(tests.test_imports.importa.importb.add(4, 5), 9)
        self.assertEqual(tests.test_imports.importa.importb.sub(4, 5), -1)
        self.assertRaises(AttributeError, lambda: tests.test_imports.importa.importb.mul(1, 2))

    def test_import_class(self):
        p_open = import_class('subprocess.Popen')
        self.assertIs(p_open, subprocess.Popen)
        self.assertRaises(ImportError, lambda: import_class('subprocess.DoesNotExist'))



BASE_LOOKUP_TABLE = {}
PLAIN_ENTRIES = set()


def register_custom_descriptor(name: str, is_plain: bool = True):
    """
    A decorator used for registering custom descriptors in order to be loadable via
    descriptor_from_dict

    Use like:

    >>> @register_custom_descriptor('ipv6')
    >>> class IPv6(Regexp):
    >>>     REGEXP = '(([0-9a-f]{1,4}:)' ...

    :param name: under which it is supposed to be invokable
    :param is_plain: is this a nested structure?
    """

    def inner(cls):
        global BASE_LOOKUP_TABLE, PLAIN_ENTRIES
        if is_plain:
            PLAIN_ENTRIES.add(name)
        BASE_LOOKUP_TABLE[name] = cls
        return cls

    return inner



import unittest
from queue import Queue
from threading import Thread
from time import sleep

from satella.coding import Monitor


class MonitorTest(unittest.TestCase):
    def test_synchronize_on(self):
        class TestedMasterClass(Monitor):
            def __init__(self):
                self.value = 0
                super().__init__()

            def get_locking_class(self):
                class LockingClass:
                    @Monitor.synchronize_on(self)
                    def get_value(self2):
                        self.value += 1

                return LockingClass()

        msc = TestedMasterClass()
        lc = msc.get_locking_class()

        class TesterThread(Thread):
            def run(self):
                lc.get_value()

        with Monitor.acquire(msc):
            TesterThread().start()
            sleep(0.1)
            self.assertEqual(msc.value, 0)
            with Monitor.release(msc):
                sleep(0.1)
                self.assertEqual(msc.value, 1)

    def test_release_contextmanager(self):
        class TestedClass(Monitor):
            def __init__(self, cqueue):
                self.cqueue = cqueue
                Monitor.__init__(self)

            @Monitor.synchronized
            def execute(self):
                self.cqueue.put(1)
                sleep(1)
                self.cqueue.get()

        class TesterThread(Thread):
            def __init__(self, tc):
                self.tc = tc
                Thread.__init__(self)

            def run(self):
                self.tc.execute()

        cq = Queue()
        cq.put(1)
        tc = TestedClass(cq)
        tt = TesterThread(tc)

        with Monitor.acquire(tc):
            with Monitor.release(tc):
                tt.start()
                sleep(0.4)
                self.assertEqual(cq.qsize(), 2)

    def test_release_contextmanager_syntax(self):
        class TestedClass(Monitor):
            def __init__(self, cqueue):
                self.cqueue = cqueue
                Monitor.__init__(self)

            @Monitor.synchronized
            def execute(self):
                self.cqueue.put(1)
                sleep(1)
                self.cqueue.get()

        class TesterThread(Thread):
            def __init__(self, tc):
                self.tc = tc
                Thread.__init__(self)

            def run(self):
                self.tc.execute()

        cq = Queue()
        cq.put(1)
        tc = TestedClass(cq)
        tt = TesterThread(tc)

        with tc:
            with Monitor.release(tc):
                tt.start()
                sleep(0.4)
                self.assertEqual(cq.qsize(), 2)

    def test_acquire_contextmanager(self):
        class TestedClass(Monitor):
            def __init__(self, cqueue):
                self.cqueue = cqueue
                Monitor.__init__(self)

            @Monitor.synchronized
            def execute(self):
                self.cqueue.put(1)
                sleep(1)
                self.cqueue.get()

        class TesterThread(Thread):
            def __init__(self, tc):
                self.tc = tc
                Thread.__init__(self)

            def run(self):
                self.tc.execute()

        cq = Queue()
        cq.put(1)
        tc = TestedClass(cq)
        tt = TesterThread(tc)

        with Monitor.acquire(tc):
            tt.start()
            sleep(0.4)
            self.assertEqual(cq.qsize(), 1)

    def test_monitoring(self):
        class TestedClass(Monitor):
            def __init__(self, cqueue):
                self.cqueue = cqueue
                Monitor.__init__(self)

            @Monitor.synchronized
            def execute(self):
                self.cqueue.put(1)
                sleep(1)
                self.cqueue.get()

        class TesterThread(Thread):
            def __init__(self, tc):
                self.tc = tc
                Thread.__init__(self)

            def run(self):
                self.tc.execute()

        q = Queue()
        tc = TestedClass(q)
        a, b = TesterThread(tc), TesterThread(tc)
        a.start(), b.start()

        while a.is_alive() or b.is_alive():
            sleep(0.1)
            self.assertNotEqual(q.qsize(), 2)

    def test_monitoring_synchronize_on_attribute(self):
        class TestedClass:
            def __init__(self, cqueue):
                self.cqueue = cqueue
                self.monitor = Monitor()

            @Monitor.synchronize_on_attribute('monitor')
            def execute(self):
                self.cqueue.put(1)
                sleep(1)
                self.cqueue.get()

        class TesterThread(Thread):
            def __init__(self, tc):
                self.tc = tc
                Thread.__init__(self)

            def run(self):
                self.tc.execute()

        q = Queue()
        tc = TestedClass(q)
        a, b = TesterThread(tc), TesterThread(tc)
        a.start(), b.start()

        while a.is_alive() or b.is_alive():
            sleep(0.1)
            self.assertNotEqual(q.qsize(), 2)



import unittest

from satella.configuration.sources import JSONSource, OptionalSource, \
    FORMAT_SOURCES
from .utils import SourceTestCase


class TestJSONSource(SourceTestCase):
    @unittest.skipIf('TOMLSource' not in FORMAT_SOURCES, 'No TOML installed')
    def test_toml(self):
        from satella.configuration.sources import TOMLSource
        self.assertSourceHas(TOMLSource(u"key = 5"), {"key": 5})
        self.assertSourceFails(TOMLSource(u'this isnt toml'))

    @unittest.skipIf('YAMLSource' not in FORMAT_SOURCES, 'No YAML installed')
    def test_yaml(self):
        from satella.configuration.sources import YAMLSource
        self.assertSourceHas(YAMLSource(u"\na: 5\n"), {"a": 5})
        self.assertSourceFails(
            YAMLSource(u'!3 #23= 43$ 434 #34 = this isnt yaml'))

    def test_ok(self):
        self.assertSourceHas(JSONSource(u'{"a":2}'), {"a": 2})

    def test_malform(self):
        self.assertSourceFails(JSONSource(u'sdfsf>>::'))
        self.assertSourceEmpty(OptionalSource(JSONSource(u'sdfsf>>::')))

    def test_base64_bad(self):
        self.assertSourceFails(JSONSource(b'bG9s\x02', encoding='base64'))

    def test_unicode_bad(self):
        self.assertSourceFails(JSONSource(b'\x00'))



import math
import unittest

from satella.coding.structures import Proxy


class TestProxy(unittest.TestCase):
    def test_proxy(self):
        a = Proxy(5.25, True)
        self.assertIsInstance(a+2, Proxy)
        a = Proxy(5.25)
        self.assertNotIsInstance(a+2, Proxy)
        self.assertEqual(int(a), 5)
        self.assertEqual(float(a), 5.25)
        self.assertEqual(a - 0.25, 5.0)
        self.assertEqual(a * 2, 10.5)
        self.assertEqual(a / 2, 2.625)
        self.assertEqual(a // 2, 2)
        self.assertEqual(a ** 2, 5.25*5.25)
        self.assertEqual(-a, -5.25)
        self.assertEqual(math.floor(a), 5)
        self.assertEqual(divmod(a, 2), (2.0, 1.25))
        self.assertEqual(divmod(2, a), (0.0, 2.0))
        self.assertEqual(round(a), 5)
        self.assertEqual(math.trunc(a), 5)
        self.assertEqual(- 0.25+a, 5.0)
        self.assertEqual(2 * a, 10.5)
        self.assertEqual(2 / a, 2/5.25)
        self.assertEqual(2 // a, 0)
        self.assertEqual(complex(a), 5.25+0j)
        self.assertEqual(2 ** a,  2**5.25)
        a += 1.0
        self.assertEqual(a, 6.25)
        a -= 1.0
        self.assertEqual(a, 5.25)
        a *= 2.0
        self.assertEqual(a, 10.5)
        a /= 2.0
        self.assertEqual(a, 5.25)
        a //= 2.0
        self.assertEqual(a, 2.0)
        a = Proxy(2)
        self.assertEqual(str(a), '2')
        self.assertEqual(repr(a), '2')
        self.assertEqual(~a, -3)
        self.assertEqual(+a, 2)
        self.assertTrue(a < 3)
        self.assertTrue(a <= 3)
        self.assertFalse(a > 3)
        self.assertNotEqual(a, 3)
        self.assertFalse(a >= 3)
        self.assertEqual(a & 2, 2)
        self.assertEqual(a | 1, 3)
        self.assertEqual(2 ^ a, 0)
        self.assertEqual(a << 1, 4)
        self.assertEqual(a >> 1, 1)
        self.assertNotEqual(a, 6)
        self.assertEqual(2 & a, 2)
        self.assertEqual(1 | a, 3)
        self.assertEqual(1 << a, 4)
        self.assertEqual(1 >> a, 0)
        self.assertEqual(hex(a), '0x2')
        self.assertEqual(oct(a), '0o2')
        a <<= 2
        self.assertEqual(a, 8)
        a >>= 2
        self.assertEqual(a, 2)
        a |= 1
        self.assertEqual(a, 3)
        a &= 2
        self.assertEqual(a, 2)
        a ^= 2
        self.assertEqual(a, 0)
        a **= 2
        self.assertEqual(a, 0)

    def test_dict(self):
        a = Proxy(dict())
        a['test'] = 2
        self.assertEqual(['test'], list(iter(a)))
        self.assertIn('test', a)
        self.assertEqual(len(a), 1)
        self.assertEqual(a['test'], 2)
        del a['test']
        self.assertRaises(KeyError, lambda: a['test'])

    def test_class(self):
        class MyClass:
            def __init__(self, test):
                self.test = test

            def __matmul__(self, other):
                return 2

            def __rmatmul__(self, other):
                return 3

        a = Proxy(MyClass(2))
        self.assertEqual(a.test, 2)
        self.assertEqual(a @ 2, 2)
        self.assertEqual(2 @ a, 3)
        self.assertIn('test', dir(a))
        del a.test
        self.assertRaises(AttributeError, lambda: a.test)




import typing as tp
from collections import deque

from satella.coding.typing import T


class PushIterable(tp.Generic[T]):
    """
    An iterable that you can add elements to it's head, and they will be popped
    before this iterable is consumed when a pop is called.

    Example:

    .. code-block:: python

        a = PushIterable([1, 2, 3])
        assert a.pop() == 1
        a.push(0)
        assert a.pop() == 0
        assert a.pop() == 2
        a.push(0)
        a.push(1)
        assert a.pop() == 1
        assert a.pop() == 0
        a.push_left(0)
        a.push_left(1)
        assert a.pop() == 0
        assert a.pop() == 1
        assert a.pop() == 3
        assertRaises(StopIteration, a.pop)
    """

    def __init__(self, iterable: tp.Iterable[T]):
        self.iterable = iter(iterable)
        self.collection = deque()

    def pop(self) -> T:
        """
        Return next element from the stack
        :return: a next element
        """
        if not self.collection:
            return next(self.iterable)
        else:
            return self.collection.pop()

    def __iter__(self):
        return self

    def __next__(self) -> T:
        return self.pop()

    def push(self, item: T) -> None:
        """
        Push an item so that next pop will retrieve this item

        :param item: item to push
        """
        self.collection.append(item)

    def push_left(self, item):
        """
        Push an item so that last pop from internal buffer will retrieve this item

        :param item: item to push
        """
        self.collection.appendleft(item)



import io
import pickle
import sys
import unittest

from satella.instrumentation import Traceback


class TestTraceback(unittest.TestCase):
    def test_no_exc(self):
        tb = Traceback()
        byte = io.BytesIO()
        byte2 = tb.pickle()
        tb.pickle_to(byte)
        byte.seek(0)
        tb2 = Traceback.from_pickle(byte)
        tb3 = Traceback.from_pickle(byte2)
        self.assertEqual(tb, tb2)
        self.assertEqual(tb2, tb3)

    def test_json(self):
        try:
            raise ValueError(u'hello')
        except ValueError:
            tb = Traceback()

        js = tb.to_json()
        self.assertEqual(tb, Traceback.from_json(js))

    def test_unserialize_error(self):
        a = 'test'
        b = pickle.dumps(a)
        self.assertRaises(ValueError, lambda: Traceback.from_pickle(b))

    def test_tb(self):

        try:
            loc = u'hello world'
            raise ValueError(u'hello')
        except ValueError:
            tb = Traceback()

            p_fmt = tb.pretty_format()
        else:
            self.fail('exception not raised')

        self.assertTrue(p_fmt)

    def test_issue_21(self):
        try:
            loc = u'hello world'
            raise ValueError(u'hello')
        except ValueError:
            tb = Traceback()
            a = tb.pickle()
            self.assertIsInstance(pickle.loads(a), Traceback)

    def test_normal_stack_frames(self):
        tb = Traceback(list(sys._current_frames().values())[0])
        tb.pretty_format()

    def test_compression_happens(self):

        try:
            loc = ' ' * (10 * 1024 * 1024)
            raise ValueError('hello')
        except ValueError:
            tb = Traceback()

        self.assertLess(len(pickle.dumps(tb, -1)), 9 * 1024 * 1024)



import unittest

from satella.coding.concurrent import LockedDataset
from satella.coding.concurrent.locked_dataset import _get_internal
from satella.coding.structures import Singleton
from satella.exceptions import WouldWaitMore, ResourceNotLocked


class TestLockedDataset(unittest.TestCase):
    def test_locked_dataset(self):
        class MyDataset(LockedDataset):

            def __init__(self):
                super(MyDataset, self).__init__()
                with self:
                    self.counter = 0

            @LockedDataset.locked(blocking=False)
            def increment(self):
                self.counter += 1

        a = MyDataset()
        with a:
            self.assertRaises(WouldWaitMore,
                              lambda: a(blocking=False).__enter__())
            self.assertRaises(WouldWaitMore, lambda: a.increment())

        a.increment()
        with a:
            self.assertEqual(a.counter, 1)

        self.assertRaises(ResourceNotLocked, lambda: a.counter)
        _get_internal(a)

    def test_locked_dataset_singleton(self):
        @Singleton
        class MyDataset(LockedDataset):

            def __init__(self):
                super(MyDataset, self).__init__()
                with self:
                    self.counter = 0

            @LockedDataset.locked(blocking=False)
            def increment(self):
                self.counter += 1

        with MyDataset():
            self.assertRaises(WouldWaitMore,
                              lambda: MyDataset()(blocking=False).__enter__())
            self.assertRaises(WouldWaitMore, lambda: MyDataset().increment())

        MyDataset().increment()
        with MyDataset():
            self.assertEqual(MyDataset().counter, 1)

        self.assertRaises(ResourceNotLocked, lambda: MyDataset().counter)



import unittest

from satella.os import whereis


class TestWhereis(unittest.TestCase):
    def test_whereis_python(self):
        """
        Note that it will fail on Windows unless you have PYTHON.EXE in your path
        """
        execs = list(whereis('python'))
        self.assertGreaterEqual(len(execs), 1)



import threading
import typing as tp
import warnings
from concurrent.futures import Executor

from satella.coding.decorators.decorators import wraps
from satella.warnings import ExperimentalWarning

local_ee = threading.local()

__all__ = ['Call', 'CallIf', 'CallWithArgumentSet', 'ExecutionEnvironment', 'call_with_ee',
           'package_for_execution', 'current_call', 'current_args', 'current_history',
           'current_kwargs', 'current_ee']


def push_call_stack(cs: tp.Callable, args: tuple = (), kwargs: tp.Optional[dict] = None) -> None:
    kwargs = kwargs or {}
    arg_tuple = ((cs, args, kwargs),)
    if not hasattr(local_ee, 'cs'):
        local_ee.cs = arg_tuple
    else:
        local_ee.cs = local_ee.cs + arg_tuple


def pop_call_stack() -> tp.Optional[tp.Tuple[tp.Callable, tuple, dict]]:
    if not hasattr(local_ee, 'cs'):
        return None
    cs = local_ee.cs
    if not cs:
        return None
    v = cs[-1]
    local_ee.cs = local_ee.cs[:-1]
    return v


def before_call(fun):
    @wraps(fun)
    def inner(self, *args, **kwargs):
        if not hasattr(local_ee, 'ee'):
            v = call_with_ee(fun, ExecutionEnvironment([(args, kwargs)]))
            return v(self, *args, **kwargs)
        else:
            return fun(self, *args, **kwargs)

    return inner


class Call:
    """
    A call to given function with a given set of arguments
    """
    __slots__ = 'fn', 'args', 'kwargs'

    def __init__(self, fn, *args, **kwargs):
        warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    @before_call
    def __call__(self, *args, **kwargs):
        """
        Call this callable.

        If an execution environment is already defined, it will be used. If not,
        a new execution environment will be defined with the 0-th set of arguments
        as args, kwargs.

        :param args: args to use as the 0-th set of arguments
        :param kwargs: kwargs to use as the 0-th set of arguments
        :return: return value
        """
        push_call_stack(self, self.args, self.kwargs)
        try:
            return self.fn(*self.args, **self.kwargs)
        finally:
            pop_call_stack()


class CallWithArgumentSet(Call):
    """
    Call a function with a set of arguments provided by the environment
    """
    __slots__ = 'fn', 'arg_set_no'

    def __init__(self, fn, arg_set_no: int = 0):
        warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
        self.fn = fn
        self.arg_set_no = arg_set_no

    @before_call
    def __call__(self, *args, **kwargs):
        try:
            ee = local_ee.ee
        except AttributeError:
            raise RuntimeError('Execution environment is required!')
        args, kwargs = ee[self.arg_set_no]
        push_call_stack(self, args, kwargs)
        try:
            return self.fn(*args, **kwargs)
        finally:
            pop_call_stack()


class CallIf(Call):
    """
    Call a function only if fn_if_call returned True
    """
    __slots__ = 'fn_to_call', 'fn_call_if'

    def __init__(self, fn_if_call: Call, fn_to_call: Call):
        warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
        self.fn_to_call = fn_to_call
        self.fn_call_if = fn_if_call

    @before_call
    def __call__(self, *args, **kwargs):
        if self.fn_call_if():
            return self.fn_to_call()


class Reduce(Call):
    """
    A call consisting of calling other calls (possibly in parallel).

    It's result will be the combination of some other function calls by a given operator,
    starting with a starting value.

    By default the starting operator just discards the results.

    :param callables: callables to call in parallel
    :param reducing_op: a callable/2 that takes previous result (or the starting value) and current
        callable result, returning a new starting value
    :param starting_value: starting value
    :param do_parallel: whether try to execute these calls in parallel, if possible.
        Parallel execution will be done only if an executor is given in the execution environment.
    """
    __slots__ = 'reducing_op', 'starting_value', 'do_parallel', 'callables'

    def __init__(self, *callables: Call,
                 reducing_op: tp.Callable[[tp.Any, tp.Any], tp.Any] = lambda a, b: None,
                 starting_value: tp.Any = 0,
                 do_parallel: bool = True):
        warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
        self.reducing_op = reducing_op
        self.starting_value = starting_value
        self.do_parallel = do_parallel
        self.callables = callables

    @before_call
    def __call__(self, *args, **kwargs):
        push_call_stack(self)
        if self.do_parallel:
            if local_ee.ee.executor is not None:
                executor = local_ee.ee.executor
                sv = self.starting_value
                futures = [executor.submit(call_with_ee(callable_, local_ee.ee, local_ee.cs))
                           for callable_ in self.callables]
                for future in futures:
                    sv = self.reducing_op(sv, future.result())
                return sv
        sv = self.starting_value
        for callable_ in self.callables:
            b = callable_()
            sv = self.reducing_op(sv, b)
        pop_call_stack()
        return sv


class ExecutionEnvironment:
    """
    This has no __slots__ so you can add anything here really
    """

    def __init__(self, argument_sets: tp.Iterable[tp.Tuple[tp.Tuple[tp.Any], tp.Dict]],
                 executor: tp.Optional[Executor] = None,
                 cs=()):
        self.arg_sets = []
        for args, kwargs in argument_sets:
            self.arg_sets.append((args, kwargs))
        self.executor = executor
        self.cs = cs

    def _set_call_stack_to(self, cs: tp.Optional[tp.List[tp.Callable]] = None):
        return ExecutionEnvironment(self.arg_sets, self.executor, cs)

    def __call__(self, callable_: Call, *args, **kwargs):
        """
        Run a given callable within the current EE.

        :param callable_: callable to run
        :return: value returned by that callable
        """
        had_cs = hasattr(local_ee, 'cs')
        if had_cs:
            prev_cs = local_ee.cs

        local_ee.cs = self.cs

        had_ee = hasattr(local_ee, 'ee')
        if had_ee:
            prev_ee = local_ee.ee
        local_ee.ee = self
        try:
            return callable_(*args, **kwargs)
        finally:
            if had_ee:
                local_ee.ee = prev_ee
            else:
                del local_ee.ee
            if had_cs:
                local_ee.cs = prev_cs
            else:
                del local_ee.cs

    def __getitem__(self, item: int) -> tp.Tuple[tp.Tuple, tp.Dict]:
        """Return the n-th argument set"""
        try:
            v = self.arg_sets[item]
        except IndexError:
            v = (), {}
        return v


def call_with_ee(callable_: tp.Callable, ee: ExecutionEnvironment,
                 _copy_call_stack_from: tp.Optional[tp.List[tp.Callable]] = None) -> tp.Callable:
    """
    Return a callable that will invoke the target callable with specified execution environment,
    but only if an EE is not defined right now.

    To explicitly provide an execution environment use this:

    >>> call_1 = Call(...)
    >>> ee = ExecutionEnvironment()
    >>> ee(call_1)

    :param callable_: callable to invoke
    :param ee: execution environment to use
    :param _copy_call_stack_from: used internally, don't use
    :return: a new callable
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)

    def inner(*args, **kwargs):
        if not hasattr(local_ee, 'ee'):
            if _copy_call_stack_from is not None:
                ef = ee._set_call_stack_to(_copy_call_stack_from)
            else:
                ef = ee
            return ef(callable_, *args, **kwargs)
        else:
            return callable_(*args, **kwargs)

    return inner


def package_for_execution(clbl: tp.Callable, ee: ExecutionEnvironment) -> tp.Callable:
    """
    Return a callable that, when called, will call specified callable in target
    execution environment.

    :param clbl: callable to run
    :param ee: EE to use
    :return: a callable
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)

    def inner():
        return ee(clbl)

    return inner


def call(fun):
    """
    Make the decorated function a callable, with it's args and kwargs to be set as a 0-th argument
    set.
    """
    a = CallWithArgumentSet(fun, 0)
    return wraps(fun)(a)


def current_ee() -> ExecutionEnvironment:
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
    return local_ee.ee


def current_call() -> Call:
    """
    Return currently processed Call, or None if not available
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
    return local_ee.cs[-1][0]


def current_args() -> tp.Optional[tuple]:
    """
    Return currently used positional arguments, or None if not available
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
    if not local_ee.cs:
        return None
    return local_ee.cs[-1][1]


def current_kwargs() -> tp.Optional[dict]:
    """
    Return currently used kwargs, or None if not available
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
    if not local_ee.cs:
        return None
    return local_ee.cs[-1][2]


def current_history() -> tp.Tuple[tp.Tuple[Call, tuple, dict]]:
    """
    Return a tuple of subsequent calls that led to this call.

    Position 0 will have the absolutely first call in the hierarchy, where -1 will have the current
    one.
    :return: a tuple of tuples (Call instance, args tuple, kwargs dict)
    """
    warnings.warn('This module is experimental, use at your own peril', ExperimentalWarning)
    return local_ee.cs



import threading
import typing as tp
from threading import Thread

from satella.exceptions import WouldWaitMore


class ThreadCollection:
    """
    A collection of threads.

    Create like:

    >>> class MyThread(Thread):
    >>>     def __init__(self, a):
    >>>         ...
    >>> tc = ThreadCollection.from_class(MyThread, [2, 4, 5], daemon=True)
    >>> tc.start()
    >>> tc.terminate()
    >>> tc.join()

    This also implements iteration (it will return all the threads in the collection) and
    length check.
    """

    __slots__ = 'threads',

    def __len__(self):
        return len(self.threads)

    def __iter__(self):
        return iter(self.threads)

    @classmethod
    def get_currently_running(cls, include_main_thread: bool = True) -> 'ThreadCollection':
        """
        Get all currently running threads as thread collection

        :param include_main_thread: whether to include the main thread

        :return: a thread collection representing all currently running threads
        """
        result = []
        for thread in threading.enumerate():
            # noinspection PyProtectedMember
            if not include_main_thread and isinstance(thread, threading._MainThread):
                continue
            result.append(thread)
        return ThreadCollection(result)

    @classmethod
    def from_class(cls, cls_to_use, iteratable, **kwargs) -> 'ThreadCollection':
        """
        Build a thread collection

        :param cls_to_use: class to instantiate with
        :param iteratable: an iterable with the sole argument to this class
        """
        return ThreadCollection([cls_to_use(it, **kwargs) for it in iteratable])

    @property
    def daemon(self) -> bool:
        """
        Is any of the threads a daemon?

        Also, when used a setter sets daemon attribute.
        """
        return any(thread.daemon for thread in self.threads)

    @daemon.setter
    def daemon(self, v: bool) -> None:
        for thread in self.threads:
            thread.daemon = v

    def __init__(self, threads: tp.Sequence[Thread]):
        self.threads = list(threads)

    def append(self, thread: Thread) -> 'ThreadCollection':
        """
        Alias for :meth:`~satella.coding.concurrent.ThreadCollection.add`

        :param thread: thread to add
        :returns: this thread collection instance
        """
        self.add(thread)
        return self

    def add(self, thread: Thread) -> 'ThreadCollection':
        """
        Add a thread to the collection

        :param thread: thread to add
        :returns: this thread collection instance
        """
        self.threads.append(thread)
        return self

    def start(self) -> 'ThreadCollection':
        """
        Start all threads

        :returns: this thread collection instance
        """
        for thread in self.threads:
            thread.start()
        return self

    def terminate(self, *args, **kwargs) -> 'ThreadCollection':
        """
        Call terminate() on all threads that have this method

        :returns: this thread collection instance
        """
        for thread in self.threads:
            try:
                thread.terminate(*args, **kwargs)
            except AttributeError:
                pass
        return self

    def join(self, timeout: tp.Optional[float] = None) -> 'ThreadCollection':
        """
        Join all threads

        :param timeout: maximum time in seconds to wait for the threads to terminate.
            Note that the timeout will be applied to each thread in sequence, so this
            can block for up to thread_count*timeout seconds. Default value of
            None means wait as long as it's necessary.
        :returns: this thread collection instance
        :raises WouldWaitMore: one of the threads failed to terminate
        """
        for thread in self.threads:
            thread.join(timeout=timeout)
            if thread.is_alive():
                raise WouldWaitMore('Thread failed to terminate')
        return self

    def is_alive(self) -> bool:
        """
        Is at least one thread alive?
        """
        return any(thread.is_alive() for thread in self.threads)



import typing as tp
import threading
import multiprocessing
import time

import psutil

from satella.coding import for_argument
from satella.coding.structures import Singleton
from satella.coding.transforms import percentile
from satella.time import parse_time_string

DEFAULT_REFRESH_EACH = '30m'
DEFAULT_WINDOW_SECONDS = '5m'


@Singleton
class CPUProfileBuilderThread(threading.Thread):
    """
    A CPU profile builder thread and a core singleton object to use.

    :param window_seconds: the amount of seconds for which to collect data.
        Generally, this should be the interval during which your system cycles through all of
        it's load, eg. if it asks it's devices each 5 minutes, the interval should be 300 seconds.
        Or a time string.
    :param refresh_each: time of seconds to sleep between rebuilding of profiles, or a time string.
    """

    def __init__(self, window_seconds: tp.Union[str, int] = DEFAULT_WINDOW_SECONDS,
                 refresh_each: tp.Union[str, int] = DEFAULT_REFRESH_EACH,
                 percentiles_requested: tp.Sequence[float] = (0.9,)):
        super().__init__(name='CPU profile builder', daemon=True)
        self.window_size = int(parse_time_string(window_seconds))
        self.refresh_each = parse_time_string(refresh_each)
        self.data = []
        self.percentiles_requested = list(percentiles_requested)
        self.percentile_values = []
        self.percentiles_regenerated = False
        self.start()

    def request_percentile(self, percent: float) -> None:
        if percent not in self.percentiles_requested:
            self.percentiles_requested.append(percent)
            self.percentiles_regenerated = False

    def percentile(self, percent: float) -> float:
        if not self.data:
            return 0
        if percent in self.percentiles_requested and self.percentiles_regenerated:
            return self.percentile_values[self.percentiles_requested.index(percent)]
        else:
            return percentile(self.data, percent)

    def is_done(self) -> bool:
        return bool(self.data)

    def recalculate(self) -> None:
        data = []
        calculate_occupancy_factor()  # as first values tend to be a bit wonky
        for _ in range(int(self.window_size)):
            time.sleep(1)
            data.append(calculate_occupancy_factor())
        percentiles = []
        data.sort()
        for percent in self.percentiles_requested:
            percentiles.append(percentile(data, percent))
        self.percentile_values = percentiles
        self.percentiles_regenerated = True
        self.data = data

    def run(self):
        while True:
            time.sleep(self.refresh_each)
            self.recalculate()


class CPUTimeManager:
    @staticmethod
    def percentile(percent: float) -> float:
        """
        Return given percentile of current CPU time's profile
        :param percent: float between 0 and 1
        :return: the value of the percentile
        """
        return CPUProfileBuilderThread().percentile(percent)

    @staticmethod
    def set_window_size(window_size: float) -> None:
        """
        Set the time that should be observed in order to build an execution profile.

        :param window_size: time, in seconds
        """
        CPUProfileBuilderThread().window_size = window_size


@for_argument(parse_time_string)
def sleep_cpu_aware(seconds: tp.Union[str, float], of_below: tp.Optional[float] = None,
                    of_above: tp.Optional[float] = None,
                    check_each: float = 1) -> bool:
    """
    Sleep for specified number of seconds.

    Quit earlier if the occupancy factor goes below of_below or above of_above
    :param seconds: time to sleep in seconds, or a time string
    :param of_below: occupancy factor below which the sleep will return
    :param of_above: occupancy factor above which the sleep will return
    :param check_each: amount of seconds to sleep at once
    :return: whether was awoken due to CPU time condition
    """
    if of_below is None and of_above is None:
        time.sleep(seconds)
        return False
    calculate_occupancy_factor()  # prime the counter
    while seconds > 0:
        time_to_sleep = min(seconds, check_each)
        time.sleep(time_to_sleep)
        of = calculate_occupancy_factor()

        if of_above is not None:
            if of > of_above:
                return True
        if of_below is not None:
            if of < of_below:
                return True
        seconds -= time_to_sleep
        if seconds <= 0:
            return False
    return False


previous_cf = None  # type: float
previous_timestamp = None  # type: float


def _calculate_occupancy_factor() -> float:
    c = psutil.cpu_times()
    try:
        try:
            try:
                used = c.user + c.nice + c.system + c.irq + c.softirq + c.steal + c.guest + c.guest_nice
            except AttributeError:
                # Linux?
                used = c.user + c.nice + c.system + c.irq + c.softirq
        except AttributeError:
            # UNIX ?
            used = c.user + c.nice + c.system
    except AttributeError:
        # windows?
        used = c.user + c.system + c.interrupt + c.dpc
    cur_time = time.monotonic()
    occupation_factor = used / multiprocessing.cpu_count()
    global previous_timestamp, previous_cf
    if previous_timestamp is None:
        previous_cf = occupation_factor
        previous_timestamp = cur_time
        return
    delta = cur_time - previous_timestamp
    if delta == 0:
        return
    of = (occupation_factor - previous_cf) / delta
    previous_cf = occupation_factor
    previous_timestamp = cur_time
    return of


def calculate_occupancy_factor() -> float:
    """
    Get the average load between now and the time it was last called as a float,
    where 0.0 is LA=0 and 1.0 is LA=max_cores.

    This will be the average between now and the time it was last called.

    .. warning:: This in rare cases (being called the first or the second time) may block for
                 up to 0.1 seconds

    :return: a float between 0 and 1 telling you how occupied CPU-wise is your system.
    """
    c = _calculate_occupancy_factor()
    while c is None:
        time.sleep(0.1)
        c = _calculate_occupancy_factor()
    return c



import collections


class DefaultDict(collections.defaultdict):
    """
    A collections.defaultdict that does not store in itself empty values that it generates
    """

    def __getitem__(self, item):
        if item not in self:
            return self.default_factory()
        return super().__getitem__(item)



import io
import logging
import unittest

from satella.instrumentation.dump_frames_on import dump_frames_on

logger = logging.getLogger(__name__)


class TestDumpFramesOn(unittest.TestCase):
    def test_dump_frames_on(self):
        io_out = io.StringIO()
        dump_frames_on(0, None, io_out)
        data = io_out.getvalue()
        self.assertIn('test_dump_frames_on', data)



import http.server
import io
import typing as tp

from satella.coding.concurrent import TerminableThread
from .. import getMetric
from ..data import MetricData, MetricDataCollection

__all__ = ['metric_data_collection_to_prometheus', 'PrometheusHTTPExporterThread']


class PrometheusHandler(http.server.BaseHTTPRequestHandler):
    """A request handler for the PrometheusHTTPExporterThread HTTP server"""

    def do_GET(self):
        """only GETs are supported"""
        if self.path != '/metrics':
            self.send_error(404, 'Unknown path. Only /metrics is supported.')
            return

        metric_data = self.server.get_metric_data()
        metric_data.add_labels(self.server.extra_labels)
        metric_data = metric_data_collection_to_prometheus(metric_data)
        self.send_response(200)
        self.send_header('Content-Type', 'text/plain; charset=utf-8')
        self.end_headers()
        self.wfile.write(metric_data.encode('utf8'))
        self.server.metric.runtime()


class PrometheusHTTPExporterThread(TerminableThread):
    """
    A daemon thread that listens on given interface as a HTTP server, ready to serve as a connection
    point for Prometheus to scrape metrics off this service.

    This additionally (if user requests so) may export a metric called prometheus.exports_per_time
    which is a cps with time_unit_vectors=[1, 20, 60] counting the amount of exports in given time
    period.

    :param interface: a interface to bind to
    :param port: a port to bind to
    :param extra_labels: extra labels to add to each metric data point, such as the name of the
        service or the hostname
    :param enable_metric: whether to enable the metric
    """

    def __init__(self, interface: str, port: int, extra_labels: tp.Optional[dict] = None,
                 enable_metric: bool = False):
        super().__init__(daemon=True)
        self.interface = interface  # type: str
        self.port = port  # type: int

        self_2 = self

        class HTTPServer(http.server.HTTPServer):
            def __init__(self):
                super().__init__((self_2.interface, self_2.port), PrometheusHandler,
                                 bind_and_activate=False)

            def get_metric_data(self):
                return self_2.get_metric_data()

        self.httpd = HTTPServer()
        self.httpd.extra_labels = extra_labels or {}
        self.httpd.metric = getMetric('prometheus.exports_per_time',
                                      'cps' if enable_metric else 'empty',
                                      time_unit_vector=[1, 20, 60])

    def run(self) -> None:
        self.httpd.server_bind()
        self.httpd.server_activate()
        self.httpd.serve_forever()
        self.httpd.server_close()

    def get_metric_data(self) -> MetricDataCollection:
        """
        Obtain metric data.

        Overload to provide custom source of metric data.
        """
        return getMetric().to_metric_data()

    def terminate(self, force: bool = False) -> 'PrometheusHTTPExporterThread':
        """
        Order this thread to terminate and return self.

        You will need to .join() on this thread to ensure that it has quit.

        :param force: whether to terminate this thread by injecting an exception into it
        """
        self.httpd.shutdown()
        return super().terminate(force=force)


class RendererObject(io.StringIO):

    def render(self, md: MetricData):

        if md.internal:  # Don't output internal metrics
            return

        self.write(md.name.replace('.', '_'))
        if md.labels:
            self.write('{')
            self.write(','.join('%s="%s"' % (
                key, str(value).replace('\\', '\\\\').replace('"', '\\"')) for
                                key, value in md.labels.items()))
            self.write('}')
        self.write(' %s' % (md.value,))
        if md.timestamp is not None:
            self.write(' %s' % (int(md.timestamp * 1000),))
        self.write('\n')


def metric_data_collection_to_prometheus(mdc: MetricDataCollection) -> str:
    """
    Render the data in the form understandable by Prometheus.

    Values marked as internal will be skipped.

    :param mdc: Metric data collection to render
    :param tree: MetricDataCollection returned by the root metric (or any metric for that instance).
    :return: a string output to present to Prometheus
    """
    if not mdc.values:
        return '\n'
    obj = RendererObject()
    for value in mdc.values:
        if value.internal:
            continue
        obj.render(value)
    return obj.getvalue()



import concurrent
import typing as tp
from concurrent.futures import Future

from satella.exceptions import WouldWaitMore


class FutureCollection:
    """
    A set of futures sharing a common result, or a common exception.

    This overloads the operator + for making an union of futures. It can be used
    with either instances of :class:`~satella.coding.concurrent.FutureCollection`
    or normal futures.

    Also supports the indexing operator to get n-th future.
    """
    __slots__ = 'futures',

    def __init__(self, futures: tp.Sequence[Future] = ()):
        if not isinstance(futures, list):
            futures = list(futures)
        self.futures = futures

    def __getitem__(self, n: int) -> Future:
        return self.futures[n]

    def __add__(self, other: tp.Union['FutureCollection', Future, tp.Sequence[Future]]):
        if isinstance(other, FutureCollection):
            return FutureCollection(self.futures + other.futures)
        elif isinstance(other, tp.Sequence):
            return FutureCollection(self.futures + list(other))
        else:
            fc = self.futures[:]
            fc.append(other)
            return FutureCollection(fc)

    def add_done_callback(self, callback, only_one: bool = False) -> None:
        """
        Add a callback to a Future to be called on it's completion.

        By default, this will add the callback to all futures.

        :param callback: callback that takes the completed Future as argument
        :param only_one: callback will be added only to a single Future. False by default
        :raises IndexError: only_one was given and no Futures in collection!
        """
        if only_one:
            self.futures[0].add_done_callback(callback)
        else:
            for future in self.futures:
                future.add_done_callback(callback)

    def set_running_or_notify_cancel(self) -> bool:
        """
        Call :code:`set_running_or_notify_cancel` on the futures

        This will return True if at least one future was not cancelled
        """
        starting = False
        for future in self.futures:
            starting = starting or future.set_running_or_notify_cancel()
        return starting

    def __iadd__(self, other: tp.Union['FutureCollection', Future, tp.Sequence[Future]]):
        if isinstance(other, FutureCollection):
            self.futures.extend(other.futures)
            return self
        elif isinstance(other, tp.Sequence):
            self.futures.extend(other)
            return self
        else:
            self.futures.append(other)
            return self

    def add(self, future: Future) -> 'FutureCollection':
        """
        Add a future

        :param future: a Future to add
        :return: self
        """
        self.futures.append(future)
        return self

    def set_result(self, result) -> None:
        """
        Set a result for all futures

        :param result: result to set
        """
        for future in self.futures:
            future.set_result(result)

    def set_exception(self, exc) -> None:
        """
        Set an exception for all futures

        :param exc: exception instance to set
        """
        for future in self.futures:
            future.set_exception(exc)

    def result(self, timeout: tp.Optional[float] = None) -> list:
        """
        Return the result of all futures, as a list.

        This will block until the results are available.

        :param timeout: a timeout in seconds for a single result. Default value None means
            wait as long as necessary
        :return: list containing results of all futures
        :raises WouldWaitMore: timeout while waiting for result
        """
        try:
            return [fut.result(timeout) for fut in self.futures]
        except concurrent.futures.TimeoutError:
            raise WouldWaitMore('timeout waiting for the result')

    def exception(self, timeout: tp.Optional[float] = None) -> tp.Optional[Exception]:
        """
        Return first exception raised by any of the futures

        This will block until the results are available.
        This call proceeding does not mean that results for all are available, since
        this will return the first exception encountered!

        :param timeout: a timeout in seconds for a single result. Default value None means
            wait as long as necessary
        :return: the first exception, or None if there were no exceptions
        :raises WouldWaitMore: timeout while waiting for result
        """
        try:
            for fut in self.futures:
                e = fut.exception(timeout)
                if e is not None:
                    return e
            return None
        except concurrent.futures.TimeoutError:
            raise WouldWaitMore('timeout waiting for the result')

    def cancel(self) -> bool:
        """
        Cancel all futures

        :return: True if all sections were cancelled
        """
        all_cancelled = True
        for future in self.futures:
            all_cancelled = all_cancelled and future.cancel()
        return all_cancelled




import multiprocessing
import os
import sys
import unittest

from mock import patch, Mock

from satella.exceptions import ResourceLocked
from satella.os import PIDFileLock
from satella import posix   # test importing

__all__ = ['posix']     # so that PyCharm doesn't complain


def acquire_lock_file_and_wait_for_signal(q, p):
    with PIDFileLock('lock', '.'):
        p.put(None)
        q.get()


class TestPidlock(unittest.TestCase):

    def test_pidlock(self):
        with PIDFileLock('lock', '.'):
            self.assertTrue(os.path.exists('./lock'))
            r = open('./lock', 'rb').read()
            try:
                r = int(r)
            except ValueError:
                return  # lol wut
            self.assertEqual(int(r), os.getpid())

        self.assertTrue(not os.path.exists('./lock'))

    def test_pidlock_multiacquire(self):
        q, p = multiprocessing.Queue(), multiprocessing.Queue()
        process = multiprocessing.Process(target=acquire_lock_file_and_wait_for_signal, args=(q, p))
        process.start()
        p.get()
        n = PIDFileLock('lock', '.')
        try:
            self.assertRaises(ResourceLocked, lambda: n.acquire())
        finally:
            q.put(None)
            process.terminate()
            process.join()


class TestDaemon(unittest.TestCase):
    @unittest.skipIf('win' in sys.platform, 'Running on Windows')
    def test_daemonize(self):
        with patch('sys.stdin') as stdin, patch('sys.stdout') as stdout, patch(
                'sys.stderr') as stderr, \
                patch('os.fork', return_value=0) as fork, patch('os.umask') as umask, patch(
            'os.setsid') as setsid, \
                patch('os.chdir') as chdir, patch('sys.exit', new=lambda: 0) as exit:
            from satella.os import daemonize

            stdin.close, stdout.close, stderr.close = Mock(), Mock(), Mock()

            daemonize()

            stdin.close.assert_called()
            stdout.close.assert_called()
            stderr.close.assert_called()
            fork.assert_called()
            setsid.assert_called()
            umask.assert_called_with(0)
            chdir.assert_called_with('/')



import logging
import typing as tp
from satella.instrumentation.memory import MemoryPressureManager, CustomCondition, All, Any
import time
import unittest
logger = logging.getLogger(__name__)


class OnDemandCondition(CustomCondition):
    def __init__(self):
        self.value = False
        super().__init__(lambda: self.value)

    def can_fire(self, *args) -> bool:
        return self.value


class TestMemory(unittest.TestCase):
    def test_memory(self):
        odc = OnDemandCondition()

        a = {'memory': False,
             'calls': 0,
             'improved': False,
             'times_entered_1': 0,
             'level_2_engaged': False,
             'level_2_confirmed': False,
             'cancelled': 0,
             'mem_normal': 0}

        cc = CustomCondition(lambda: a['level_2_engaged'])

        MemoryPressureManager(None, [odc, All(cc, Any(cc, cc))], 2)

        def memory_normal():
            nonlocal a
            a['mem_normal'] += 1

        def cancel():
            nonlocal a
            a['cancelled'] += 1
        MemoryPressureManager.register_on_memory_normal(memory_normal)

        cc = MemoryPressureManager.register_on_entered_severity(1)(cancel)

        @MemoryPressureManager.register_on_entered_severity(2)
        def call_on_level_2():
            a['level_2_confirmed'] = True

        @MemoryPressureManager.register_on_remaining_in_severity(1)
        def call_on_memory_still():
            a['calls'] += 1

        @MemoryPressureManager.register_on_entered_severity(1)
        def call_on_no_memory():
            a['memory'] = True
            a['times_entered_1'] += 1

        @MemoryPressureManager.register_on_left_severity(1)
        def call_improved():
            a['improved'] = True

        self.assertFalse(a['memory'])
        self.assertFalse(a['improved'])
        self.assertEqual(a['mem_normal'], 0)
        time.sleep(3)
        odc.value = True
        time.sleep(5)
        self.assertEqual(a['cancelled'], 1)
        cc.cancel()
        self.assertTrue(a['memory'])
        self.assertFalse(a['improved'])
        self.assertGreater(a['calls'], 0)
        self.assertEqual(a['times_entered_1'], 1)
        odc.value = False
        time.sleep(3)
        self.assertTrue(a['improved'])
        self.assertEqual(a['times_entered_1'], 1)
        self.assertTrue(a['memory'])
        self.assertEqual(a['mem_normal'], 1)
        a['level_2_engaged'] = True
        time.sleep(3)
        self.assertEqual(MemoryPressureManager().severity_level, 2)
        self.assertEqual(a['cancelled'], 1)
        self.assertEqual(a['times_entered_1'], 2)
        self.assertTrue(a['level_2_confirmed'])
        self.assertEqual(a['mem_normal'], 1)



import unittest

from satella.coding import static_var


class FunTestTest(unittest.TestCase):
    def test_fun_static_function(self):
        @static_var("counter", 2)
        def static_fun(a):
            static_fun.counter += 1
            return a

        static_fun(2)
        static_fun(3)
        self.assertEqual(static_fun.counter, 4)

    def test_fun_static_method(self):
        class MyClass(object):
            @static_var("counter", 2)
            def my_method(self):
                MyClass.my_method.counter += 1
                return a

        a = MyClass()
        a.my_method()
        a.my_method()
        self.assertEqual(MyClass.my_method.counter, 4)



import logging
import unittest

from satella.coding import rethrow_as, silence_excs, catch_exception, log_exceptions, \
    raises_exception, reraise_as

logger = logging.getLogger(__name__)


class TestStuff(unittest.TestCase):

    def test_raises(self):
        self.assertTrue(raises_exception(NameError, lambda: name_error))
        self.assertFalse(raises_exception(NameError, lambda: None))

    def test_silence_excs_returns_factory(self):
        try:
            @silence_excs(KeyError, returns=5, returns_factory=5)
            def errors():
                pass
        except ValueError:
            pass
        else:
            self.fail('ValueError not raised')

        @silence_excs(KeyError, returns_factory=lambda: 5)
        def key_error():
            raise KeyError()

        self.assertEqual(key_error(), 5)

    def test_log_exceptions_swallow(self):

        @log_exceptions(logger, logging.CRITICAL, exc_types=KeyError, swallow_exception=True)
        def get_me(exc_type):
            yield 2
            yield 4
            raise exc_type()

        @log_exceptions(logger, logging.CRITICAL, exc_types=KeyError, swallow_exception=True)
        def get_me_normal(exc_type):
            raise exc_type()

        for i in get_me(KeyError):
            pass
        get_me_normal(KeyError)

        self.assertRaises(ValueError, lambda: list(get_me(ValueError)))
        self.assertRaises(ValueError, lambda: get_me_normal(ValueError))

    def test_log_exceptions_decorator(self):

        @log_exceptions(logger, logging.CRITICAL)
        def get_me():
            yield 2
            yield 4
            raise KeyError()

        try:
            with log_exceptions(logger, exc_types=KeyError):
                for i in get_me():
                    pass
        except KeyError:
            pass
        else:
            self.fail('KeyError not raised')

    def test_log_exceptions(self):
        a = 5

        @silence_excs(KeyError)
        @log_exceptions(logger, logging.CRITICAL, '{a} {e} {args}', locals())
        def raise_keyerror():
            raise KeyError('hello')

        raise_keyerror()

        with silence_excs(KeyError):
            with log_exceptions(logger):
                raise_keyerror()

    def test_returns(self):
        @silence_excs(ValueError, returns=5)
        def throw_value_error():
            raise ValueError()

        self.assertEqual(5, throw_value_error())

    def test_catching(self):
        def throw_value_error():
            raise ValueError()

        def nop():
            return 4

        self.assertRaises(ValueError, lambda: catch_exception(TypeError, throw_value_error))
        self.assertIsInstance(catch_exception(ValueError, throw_value_error), ValueError)
        self.assertRaises(ValueError, lambda: catch_exception(ValueError, nop))
        self.assertEqual(4, catch_exception(TypeError, nop, return_value_on_no_exception=True))
        self.assertEqual(5, catch_exception(TypeError, nop, return_instead=5))

        self.assertIsInstance(catch_exception((ValueError, TypeError), throw_value_error),
                              ValueError)

    @silence_excs(ValueError)
    def test_silencer_2(self):
        raise ValueError()

    def test_silencer(self):
        with silence_excs(TypeError):
            raise TypeError()

    def test_rethrow(self):
        try:
            with rethrow_as((ValueError, NameError)):
                raise ValueError()
        except NameError:
            return

        self.fail()

    def test_rethrow_3(self):
        try:
            with rethrow_as(ValueError, NameError):
                raise ValueError()
        except NameError:
            return

        self.fail()

    def test_rethrow_2(self):

        @rethrow_as((ValueError, NameError))
        def lol():
            raise ValueError()

        self.assertRaises(NameError, lol)

    def test_reraise(self):
        try:
            with reraise_as(ValueError, NameError):
                raise ValueError()
        except NameError:
            return

        self.fail()

    def test_reraise_silencer(self):

        @reraise_as(ValueError, None)
        def lol():
            raise ValueError()

        lol()

    def test_issue_10(self):

        class WTFException1(Exception): pass

        class WTFException2(Exception): pass

        @rethrow_as((NameError, WTFException1),
                    (TypeError, WTFException2))
        def provide(exc):
            raise exc()

        self.assertRaises(WTFException1, lambda: provide(NameError))
        self.assertRaises(WTFException2, lambda: provide(TypeError))

    def test_issue_14a(self):

        @rethrow_as(*(((NameError, ValueError), TypeError),))
        def ro(p):
            raise p()

        self.assertRaises(TypeError, lambda: ro(NameError))
        self.assertRaises(TypeError, lambda: ro(ValueError))
        self.assertRaises(RuntimeError, lambda: ro(RuntimeError))

    def test_issue_14b(self):

        @rethrow_as((NameError, ValueError), TypeError)
        def ro(p):
            raise p()

        self.assertRaises(TypeError, lambda: ro(NameError))
        self.assertRaises(TypeError, lambda: ro(ValueError))
        self.assertRaises(RuntimeError, lambda: ro(RuntimeError))



import os

import psutil

from satella.exceptions import ResourceLocked


class PIDFileLock:
    """
    Acquire a PID lock file.

    Usage:

    >>> with PIDFileLock('my_service.pid'):
    >>>     ... rest of code ..

    Any alternatively

    >>> pid_lock = PIDFileLock('my_service.pid')
    >>> pid_lock.acquire()
    >>> ...
    >>> pid_lock.release()

    The constructor doesn't throw, __enter__ or acquire() does, one of:

   * ResourceLocked - lock is already held. This has two attributes - pid (int), the PID of holder,
                                  and is_alive (bool) - whether the holder is an alive process
    """
    __slots__ = ('path', 'file_no')

    def __init__(self, pid_file, base_dir=u'/var/run'):
        """
        Initialize a PID lock file object

        :param pid_file: rest of path
        :param base_dir: base lock directory
        """
        self.path = os.path.join(base_dir, pid_file)
        self.file_no = None

    def release(self) -> None:
        """
        Free the lock
        :raises RuntimeError: lock not acquired
        """
        if self.file_no is None:
            raise RuntimeError('lock not acquired')
        os.unlink(self.path)
        self.file_no = None

    def acquire(self) -> None:
        """
        Acquire the PID lock

        :raises ResourceLocked: if lock if held
        """
        try:
            self.file_no = os.open(self.path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        except (OSError, FileExistsError):
            with open(self.path, 'r') as fin:
                data = fin.read().strip()

            try:
                pid = int(data)
            except ValueError:
                os.unlink(self.path)
                return self.acquire()

            if pid in {x.pid for x in psutil.process_iter()}:
                raise ResourceLocked(pid)
            else:
                # does not exist
                os.unlink(self.path)
                return self.acquire()

        fd = os.fdopen(self.file_no, 'w')
        fd.write(str(os.getpid()) + '\n')
        fd.close()

    def __enter__(self) -> None:
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



import unittest

from satella.coding import precondition, short_none, has_keys, update_if_not_none, postcondition, \
    get_arguments, expect_exception
from satella.coding.decorators import for_argument
from satella.exceptions import PreconditionError


class TestTypecheck(unittest.TestCase):

    def test_except_exception(self):
        def expect_exception_1():
            with expect_exception(KeyError, ValueError, 'Hello world!'):
                pass

        self.assertRaises(ValueError, expect_exception_1)

        a = {}
        with expect_exception(KeyError, ValueError, 'Hello world!'):
            a['test']

        def expect_exception_2():
            with expect_exception(KeyError, ValueError, 'Hello world!'):
                raise TypeError()

        self.assertRaises(TypeError, expect_exception_2)

    def test_for_argument_bug(self):
        class Device:
            @for_argument(None, str)
            def __init__(self, device_id: str, init_true: bool = False):
                self.device_id = device_id
                self.init_true = init_true

            def __eq__(self, other):
                return self.device_id == other.device_id

        d = Device(1234)
        self.assertEqual(d.device_id, '1234')
        self.assertFalse(d.init_true)

    def test_get_arguments_bug(self):
        class Class:

            @classmethod
            def execute(cls, password: str, service: str,
                        absolute_expiration=None,
                        metadata=None):
                pass

            def method_call(self, password: str, service: str,
                        absolute_expiration=None,
                        metadata=None):
                pass
        kls = Class()

        args = get_arguments(Class.execute, '1234',
                             'service', metadata={'ok': 'meta'})
        self.assertEqual(args, {
            'password': '1234',
            'service': 'service',
            'metadata': {'ok': 'meta'},
            'absolute_expiration': None
        })

        args = get_arguments(kls.method_call, '1234', 'service', metadata={'ok': 'meta'})
        self.assertEqual(args, {
            'password': '1234',
            'service': 'service',
            'metadata': {'ok': 'meta'},
            'absolute_expiration': None
        })

        self.assertRaises(TypeError, lambda: get_arguments(kls.method_call, '1234'))
        self.assertRaises(TypeError, lambda: get_arguments(Class.execute, '1234'))

    def test_for_argument_extended(self):
        @for_argument(int)
        def accept_default(a='5'):
            return a

        self.assertEqual(accept_default('6'), 6)
        self.assertEqual(accept_default(), 5)

    def test_for_argument_str(self):
        @for_argument('x*2')
        def accept_two(x):
            self.assertEqual(x, 2)
        accept_two(1)

    def test_postcondition(self):
        @postcondition(lambda x: x)
        def return_nothing():
            pass

        return_nothing()

    def test_postcondition_2(self):
        @postcondition('x')
        def return_nothing():
            pass

        return_nothing()

    def test_postcondition_3(self):
        @postcondition(lambda x: False)
        def return_nothing():
            pass

        self.assertRaises(PreconditionError, return_nothing)

    def test_update_if_not_none(self):
        a = {}
        update_if_not_none(a, 'k', 2)
        self.assertEqual(a['k'], 2)

    def test_has_keys(self):
        p = has_keys(['a', 'b'])
        self.assertRaises(PreconditionError, lambda: p({'a': 3}))
        p({'a': 3, 'b': 4})

    def test_short_none(self):
        a = short_none('x == 2')
        self.assertTrue(a(2))
        self.assertIsNone(a(None))
        self.assertFalse(a(3))

    def test_precondition_none(self):
        @precondition(a=None)
        def do_any(a):
            return a

        self.assertEqual(do_any(a=5), 5)

    def test_precondition_kwargs(self):
        @precondition(value='x == 2')
        def kwarg(**kwargs):
            return kwargs['value']

        self.assertRaises(PreconditionError, lambda: kwarg(value=3))
        kwarg(value=2)

    def test_precondition(self):
        @precondition('len(x) == 1', lambda x: x == 1, None)
        def return_double(x, y, z):
            pass

        self.assertRaises(PreconditionError, lambda: return_double([], 1, 5))
        self.assertRaises(PreconditionError, lambda: return_double([1], 2, 5))
        return_double([1], 1, 'dupa')

    def test_precondition_fail(self):
        @precondition('len(x) == 1', lambda x: x == 1, None, None, None)
        def return_double(x, y, z):
            pass

        self.assertRaises(AssertionError, lambda: return_double([1], 1, 5))

    def test_forarg_kwords(self):
        @for_argument(int, typed=bool)
        def check(v1, typed='True'):
            if typed:
                return v1

        self.assertEqual(check('5'), 5)

    def test_forarg_shorter(self):
        @for_argument(int)
        def check(v1, v2):
            if v2 != 'str':
                raise ValueError()
            return v1

        self.assertEqual(check('5', 'str'), 5)

    def test_forarg_returns(self):
        @for_argument(returns=bool)
        def check(v1, v2):
            return v1 and v2

        self.assertTrue(check('dupa', 'zbita'))

    def test_forarg(self):
        @for_argument(int)
        def testa(a):
            return a

        self.assertEqual(testa('5'), 5)
        self.assertEqual(testa(5), 5)
        self.assertIsInstance(testa('5'), int)



from abc import ABCMeta, abstractmethod

from satella.coding.decorators.decorators import wraps

__all__ = ['Loadable', 'must_be_loaded']


class Loadable(metaclass=ABCMeta):
    """
    Any class that can be loaded lazily.

    It's keyword argument, load_lazy is expected to control lazy loading. If set to True,
    DB will be hit as a part of this object's constructor.

    If False, you will need to load it on-demand via :func:`must_be_loaded` decorator.
    """

    __slots__ = '_loaded',

    def __init__(self, load_lazy: bool = False):
        self._loaded = False
        if not load_lazy:
            self.refresh()

    @abstractmethod
    def refresh(self, load_from=None) -> None:
        """
        Ask the database about this object, or load it from provided serialized representation.

        Override me, calling me in a super method.

        :param load_from: serialized object. If not given, the DB will be asked for it
        """
        self._loaded = True


def must_be_loaded(fun):
    """
    A decorator for :class:`Loadable`'s methods.

    Assures that :meth:`~Loadable.refresh` is called prior to executing that method, ie. the object
    is loaded from the DB.
    """

    @wraps(fun)
    def inner(self, *args, **kwargs):
        assert isinstance(self,
                          Loadable), 'must_be_loaded called with a class that does not subclass ' \
                                     'Loadable'
        if not self._loaded:
            self.refresh()
        return fun(self, *args, **kwargs)

    return inner



import queue
import typing as tp
import warnings

__all__ = ['BaseSatellaError', 'ResourceLockingError', 'ResourceNotLocked', 'ResourceLocked',
           'ConfigurationValidationError', 'ConfigurationError', 'ConfigurationSchemaError',
           'PreconditionError', 'MetricAlreadyExists', 'BaseSatellaException', 'CustomException',
           'CodedCustomException', 'CodedCustomExceptionMetaclass', 'WouldWaitMore',
           'ProcessFailed', 'AlreadyAllocated', 'Empty', 'ImpossibleError',
           'ConfigurationMisconfiguredError']


class CustomException(Exception):
    """"
    Base class for your custom exceptions. It will:

    1. Accept any number of arguments
    2. Provide faithful __repr__ and a reasonable __str__

    It passed all arguments that your exception received via super().
    Just remember to actually pass these arguments in your inheriting classes!
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.kwargs = kwargs

    def __str__(self) -> str:
        a = '%s(%s' % (self.__class__.__qualname__.split('.')[-1], ', '.join(map(repr, self.args)))
        try:
            if self.kwargs:
                a += ', ' + ', '.join(map(lambda k, v: '%s=%s' % (k, repr(v)), self.kwargs.items()))
        except AttributeError:
            pass
        a += ')'
        return a

    def __repr__(self) -> str:
        a = '%s%s(%s' % ((self.__class__.__module__ + '.')
                         if self.__class__.__module__ != 'builtins' else '',
                         self.__class__.__qualname__,
                         ', '.join(map(repr, self.args)))
        try:
            if self.kwargs:
                a += ', ' + (', '.join(map(lambda kv: '%s=%s' % (kv[0], repr(kv[1])),
                                           self.kwargs.items())))
        except AttributeError:
            pass
        a += ')'
        return a


def get_base_of_bases(classes):
    class_bases = ()
    for class_ in classes:
        class_bases += class_.__bases__
    return class_bases


class CodedCustomExceptionMetaclass(type):
    """
    Metaclass implementing the isinstance check for coded custom exceptions
    """
    code = None  # type: tp.Optional[tp.Any]

    def __instancecheck__(cls, instance):
        if super().__instancecheck__(instance):
            return True

        if cls is CodedCustomException:
            return super().__instancecheck__(instance)

        class_base = (cls,)
        while CodedCustomException not in get_base_of_bases(class_base) and class_base:
            class_base = get_base_of_bases(class_base)

        inst_base = (instance.__class__,)
        while CodedCustomException not in get_base_of_bases(inst_base) and inst_base:
            inst_base = get_base_of_bases(inst_base)

        if len(set(class_base).intersection(set(inst_base))) == 0:
            # These classes belong in different exception hierarchies
            return False

        try:
            a = cls.code == instance.code
        except AttributeError:
            a = super().__instancecheck__(instance)
        return a


class CodedCustomException(CustomException, metaclass=CodedCustomExceptionMetaclass):
    """
    An exception with the property that every CodedCustomException having the code
    of it's child instance will be considered it's child by isinstance.

    I.e. following is true:

    >>> class MyCustomError(CodedCustomException):
    >>>     code = 5
    >>> a = CodedCustomException()
    >>> a.code = 5
    >>> assert isinstance(a, MyCustomError)

    Note that you however cannot use your child classes, ie. in following way:

    >>> try:
    >>>     e = CodedCustomException()
    >>>     e.code = 5
    >>> except MyCustomError:
    >>>     ...

    due to Python limitations. See
    https://mail.python.org/pipermail/python-ideas/2015-November/037104.html to learn more
    about the issue.

    Note that the exception you're checking needs to be included in the class hierarchy.
    Here's the snippet that illustrates this:

    >>> class MyException(CodedCustomException):
    >>>     pass
    >>> class MyCode5(MyException):
    >>>     code = 5
    >>> class DifferentHierarchy(CodedCustomException):
    >>>     pass
    >>> class MyCode5Diff(DifferentHierarchy):
    >>>     code = 5
    >>> a = MyException()
    >>  a.code = 5
    >>> assert not isinstance(a, MyCode5Diff)
    """

    def __init__(self, message='', code=None, *args, **kwargs):
        super().__init__(message, code, *args, **kwargs)
        self.message = message  # type: str
        if code is not None:
            self.code = code


class BaseSatellaError(CustomException):
    """"Base class for all Satella exceptions"""


class BaseSatellaException(BaseSatellaError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn('This is deprecated and will be removed in Satella 3.0, '
                      'use BaseSatellaError instead', DeprecationWarning)


class ResourceLockingError(BaseSatellaError):
    """Base class for resource locking issues"""


class ResourceLocked(ResourceLockingError):
    """Given resource has been already locked"""


class ResourceNotLocked(ResourceLockingError):
    """Locking given resource is needed in order to access it"""


class WouldWaitMore(ResourceLockingError, TimeoutError):
    """wait()'s timeout has expired"""


class PreconditionError(BaseSatellaError, ValueError):
    """
    A precondition was not met for the argument
    """


class Empty(BaseSatellaError, queue.Empty):
    """The queue was empty"""


class ConfigurationError(BaseSatellaError, ValueError):
    """A generic error during configuration"""


class ConfigurationSchemaError(ConfigurationError):
    """Schema mismatch to what was seen"""


class ConfigurationMisconfiguredError(ConfigurationError):
    """Configuration was improperly passed to Satella"""


class ConfigurationValidationError(ConfigurationSchemaError):
    """A validator failed"""

    def __init__(self, msg, value=None, **kwargs):
        """
        :param value: value found
        """
        super().__init__(msg, value, **kwargs)
        self.value = value


class MetricAlreadyExists(BaseSatellaError):
    """Metric with given name already exists, but with a different type"""

    def __init__(self, msg, name, requested_type, existing_type):
        super().__init__(msg)
        self.name = name
        self.requested_type = requested_type
        self.existing_type = existing_type


class AlreadyAllocated(ResourceLocked):
    """
    Given ID has been already marked as allocated
    """


class ProcessFailed(BaseSatellaError, OSError):
    """
    A process finished with other result code than it was requested

    :param rc: return code of the process
    :param stdout_so_far: process' stdout gathered so far
    """

    def __init__(self, rc: int, stdout_so_far: tp.Union[bytes, str]):
        super().__init__(rc, stdout_so_far)
        self.rc = rc
        self.stdout_so_far = stdout_so_far

    def __str__(self):
        return 'ProcessFailed(%s)' % (self.rc,)


class ImpossibleError(BaseException):
    """
    For these cases where your execution flow goes some place, that should be impossible
    for it to reach.

    This is a BaseException, since it should be propagated upwards as soon as possible!
    """



from .cache_dict import CacheDict, LRUCacheDict
from .counting import CountingDict
from .dict_object import apply_dict_object, DictObject
from .expiring import ExpiringEntryDict, SelfCleaningDefaultDict
from .objects import DirtyDict, DictionaryView, KeyAwareDefaultDict, TwoWayDictionary
from .writeback_cache import ExclusiveWritebackCache
from .default import DefaultDict

__all__ = ['DictObject', 'DirtyDict', 'DictionaryView', 'CacheDict', 'KeyAwareDefaultDict',
           'TwoWayDictionary', 'apply_dict_object', 'ExpiringEntryDict',
           'SelfCleaningDefaultDict', 'ExclusiveWritebackCache',
           'CountingDict', 'LRUCacheDict', 'DefaultDict']



import typing as tp

from .metric_types import Metric, MetricLevel
from .metric_types.measurable_mixin import MeasurableMixin
from ...coding.sequences import Multirun


class AggregateMetric(Metric, MeasurableMixin):
    """
    A virtual metric grabbing a few other metrics and having a single .handle() call represent a
    bunch of calls to other metrics. Ie, the following:

    >>> m1 = getMetric('summary', 'summary')
    >>> m2 = getMetric('histogram', 'histogram')
    >>> m1.runtime()
    >>> m2.runtime()

    Is the same as:

    >>> am = AggregateMetric(getMetric('summary', 'summary'), getMetric('histogram', 'histogram'))
    >>> am.runtime()

    Note that this class supports only reporting. It doesn't read data, or read/write metric levels.
    """
    __slots__ = ('metrics',)

    def __init__(self, *metrics):
        self.metrics = metrics

    def handle(self, level: tp.Union[int, MetricLevel], *args, **kwargs) -> None:
        Multirun(self.metrics).handle(level, *args, **kwargs)



import logging
import time
import typing as tp

from satella.coding.structures import CacheDict, LRUCacheDict, ExclusiveWritebackCache
from satella.coding.typing import K, V
from .. import Metric
from ..metric_types.callable import CallableMetric
from ..metric_types.counter import CounterMetric
from ..metric_types.measurable_mixin import MeasurableMixin

logger = logging.getLogger(__name__)


class MetrifiedCacheDict(CacheDict[K, V]):
    """
    A CacheDict with metrics!

    :param cache_hits: a counter metric that will be updated with +1 each time there's a cache hit
    :param cache_miss: a counter metric that will be updated with +1 each time there's a cache miss
    :param refreshes: a metric that will be updated with +1 each time there's a cache refresh
    :param how_long_refresh_takes: a metric that will be ticked with time value_getter took
    """

    def __init__(self, stale_interval, expiration_interval, value_getter,
                 value_getter_executor=None, cache_failures_interval=None,
                 time_getter=time.monotonic,
                 default_value_factory=None,
                 cache_hits: tp.Optional[CounterMetric] = None,
                 cache_miss: tp.Optional[CounterMetric] = None,
                 refreshes: tp.Optional[CounterMetric] = None,
                 how_long_refresh_takes: tp.Optional[MeasurableMixin] = None):
        if refreshes:
            old_value_getter = value_getter

            def value_getter_replacement(item):
                try:
                    return old_value_getter(item)
                finally:
                    if self.refreshes:
                        self.refreshes.runtime(+1)

            value_getter = value_getter_replacement

        if how_long_refresh_takes:
            value_getter = how_long_refresh_takes.measure(value_getter=time_getter)(value_getter)

        super().__init__(stale_interval, expiration_interval, value_getter,
                         value_getter_executor, cache_failures_interval, time_getter,
                         default_value_factory)
        self.cache_hits = cache_hits
        self.cache_miss = cache_miss
        self.refreshes = refreshes
        self.how_long_refresh_takes = how_long_refresh_takes

    def __getitem__(self, item):
        if self.has_info_about(item):
            if self.cache_hits:
                self.cache_hits.runtime(+1)
        else:
            if self.cache_miss:
                self.cache_miss.runtime(+1)
        return super().__getitem__(item)


class MetrifiedLRUCacheDict(LRUCacheDict[K, V]):
    """
    A LRUCacheDict with metrics!

    :param cache_hits: a counter metric that will be updated with +1 each time there's a cache hit
    :param cache_miss: a counter metric that will be updated with +1 each time there's a cache miss
    :param refreshes: a metric that will be updated with +1 each time there's a cache refresh
    :param how_long_refresh_takes: a metric that will be ticked with time value_getter took
    """

    def __init__(self, stale_interval: float, expiration_interval: float,
                 value_getter, value_getter_executor=None,
                 cache_failures_interval=None,
                 time_getter=time.monotonic,
                 default_value_factory=None,
                 max_size: int = 100,
                 cache_hits: tp.Optional[Metric] = None,
                 cache_miss: tp.Optional[Metric] = None,
                 refreshes: tp.Optional[Metric] = None,
                 how_long_refresh_takes: tp.Optional[MeasurableMixin] = None,
                 evictions: tp.Optional[Metric] = None,
                 **kwargs):
        if refreshes:
            old_value_getter = value_getter

            def value_getter_replacement(item):
                try:
                    return old_value_getter(item)
                finally:
                    if self.refreshes:
                        self.refreshes.runtime(+1)

            value_getter = value_getter_replacement

        if how_long_refresh_takes:
            value_getter = how_long_refresh_takes.measure(value_getter=time_getter)(value_getter)

        super().__init__(stale_interval, expiration_interval, value_getter,
                         value_getter_executor, cache_failures_interval, time_getter,
                         default_value_factory, max_size=max_size)
        self.cache_hits = cache_hits
        self.cache_miss = cache_miss
        self.refreshes = refreshes
        self.evictions = evictions
        self.how_long_refresh_takes = how_long_refresh_takes

    def evict(self):
        self.evictions.runtime(+1)
        super().evict()

    def __getitem__(self, item):
        if self.has_info_about(item):
            if self.cache_hits:
                self.cache_hits.runtime(+1)
        else:
            if self.cache_miss:
                self.cache_miss.runtime(+1)
        return super().__getitem__(item)


class MetrifiedExclusiveWritebackCache(ExclusiveWritebackCache[K, V]):
    __slots__ = ('cache_miss', 'cache_hits')

    def __init__(self, *args,
                 cache_hits: tp.Optional[CounterMetric] = None,
                 cache_miss: tp.Optional[CounterMetric] = None,
                 entries_waiting: tp.Optional[CallableMetric] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.cache_miss = cache_miss
        self.cache_hits = cache_hits
        if entries_waiting is not None:
            entries_waiting.callable = self.get_queue_length()

    def __getitem__(self, item):
        if item in self.in_cache:
            if self.cache_hits:
                self.cache_hits.runtime(+1)
        else:
            if self.cache_miss:
                self.cache_miss.runtime(+1)
        return super().__getitem__(item)



import unittest
from concurrent.futures.thread import ThreadPoolExecutor

from satella.coding.call_hierarchy import Call, CallWithArgumentSet, Reduce, CallIf, \
    ExecutionEnvironment, package_for_execution, call, current_args, current_kwargs, current_history


class TestCallHierarchy(unittest.TestCase):
    def test_call(self):
        call_ = Call(lambda: 5)
        self.assertEqual(call_(), 5)

    def test_exec_parallel(self):
        arg_sets = []

        def mult(y):
            return y*2

        for value in [1, 2, 3, 4, 5, 6, 7]:
            arg_sets.append(((value, ), {}))

        calls = []
        for i, _ in enumerate(arg_sets):
            calls.append(CallWithArgumentSet(mult, i))
        call_v = Reduce(*calls, reducing_op=lambda a, b: a+b, starting_value=0)
        tpe = ThreadPoolExecutor(max_workers=4)

        ee = ExecutionEnvironment(arg_sets, tpe)
        pp = package_for_execution(call_v, ee)
        self.assertEqual(pp(), 56)

    def test_current(self):
        @call
        def nested(a: int):
            self.assertEqual(len(current_history()), 2)

        @call
        def call_me(a: int):
            self.assertEqual(a, 5)
            self.assertEqual(current_args(), (5, ))
            self.assertEqual(current_kwargs(), {})
            nested()

        call_me(5)

    def test_arg_sets(self):
        def add(a, b):
            return a+b

        call1 = CallWithArgumentSet(add, 0)
        call2 = CallWithArgumentSet(add, 1)
        call_v = Reduce(call1, call2, reducing_op=lambda a, b: a+b, starting_value=0)
        ee = ExecutionEnvironment([((1, 2), {}), ((3, 4), {})])
        self.assertEqual(ee(call_v), 10)

    def test_call_if(self):
        a = {'test': True, 'b': 0}

        def a_is_true():
            nonlocal a
            return a['test']

        def incr_b():
            nonlocal a
            a['b'] += 1

        call_if_true = CallIf(a_is_true, incr_b)

        call_if_true()
        self.assertEqual(a['b'], 1)
        a['test'] = False
        call_if_true()
        self.assertEqual(a['b'], 1)



import unittest
import threading
import logging
import sys

from satella.exception_handling import MemoryErrorExceptionHandler

logger = logging.getLogger(__name__)


class TestMemoryHandler(unittest.TestCase):
    def setUp(self) -> None:
        # simulate giving a CallableGroup
        self.meh = MemoryErrorExceptionHandler(custom_hook=lambda type_, value, traceback: [True])
        self.meh.install()

    def test_memory_error_exception_handler(self):

        class MyThread(threading.Thread):
            def run(self):
                raise MemoryError()

        mt = MyThread()
        mt.start()
        mt.join()
        self.assertFalse(self.meh._free_on_memory_error)
